@@ -22,6 +22,8 @@ import os
 import inspect
 import re
 import copy
+import shutil
+from pathlib import Path
 from functools import partial
 from itertools import zip_longest
 
@@ -45,6 +47,8 @@ from .monkeypatch import patch_docfields
 from .directives import RemarksDirective, TodoDirective
 from .nodes import remarks
 
+import subprocess
+from docuploader import shell
 
 class Bcolors:
     HEADER = '\033[95m'
@@ -80,7 +84,31 @@ REF_PATTERN_LAST = '~([a-zA-Z0-9_<>]*\.)*[a-zA-Z0-9_<>]*(\(\))?'
 PROPERTY = 'property'
 
 
+# Run sphinx-build with Markdown builder in the plugin.
+def run_sphinx_markdown():
+    cwd = os.getcwd()
+    # Skip running sphinx-build for Markdown for some unit tests.
+    # Not required other than to output DocFX YAML.
+    if "docs" in cwd:
+        return
+
+    return shell.run(
+        [
+            "sphinx-build",
+            "-M",
+            "markdown",
+            "docs/",
+            "docs/_build",
+        ],
+        hide_output=False
+    )
+
+
 def build_init(app):
+    print("Running sphinx-build with Markdown first...")
+    run_sphinx_markdown()
+    print("Completed running sphinx-build with Markdown files.")
+
     """
     Set up environment data
     """
@@ -101,6 +129,10 @@ def build_init(app):
     app.env.docfx_info_uid_types = {}
     # This stores uidnames of docstrings already parsed
     app.env.docfx_uid_names = {}
+    # This stores file path for class when inspect cannot retrieve file path
+    app.env.docfx_class_paths = {}
+    # This stores the name and href of the markdown pages.
+    app.env.markdown_pages = []
 
     app.env.docfx_xrefs = {}
 
@@ -327,24 +359,37 @@ def _extract_docstring_info(summary_info, summary, name):
     }
     
     initial_index = -1
+    front_tag = '<xref'
+    end_tag = '/xref>'
+    end_len = len(end_tag)
         
     # Prevent GoogleDocstring crashing on custom types and parse all xrefs to normal
-    if '<xref:' in parsed_text:
+    if front_tag in parsed_text:
         type_pairs = []
-        initial_index = max(0, parsed_text.find('<xref'))
+        # Constant length for end of xref tag
+        initial_index = max(0, parsed_text.find(front_tag))
 
         summary_part = parsed_text[initial_index:]
        
-        # Remove all occurrences of "<xref:type>"
-        while "<xref:" in summary_part:
+        # Remove all occurrences of "<xref uid="uid">text</xref>"
+        while front_tag in summary_part:
 
-            # Expecting format of "<xref:type>:"
-            if "<xref:" in summary_part:
-                initial_index += summary_part.find("<xref")
-                original_type = parsed_text[initial_index:initial_index+(parsed_text[initial_index:].find('>'))+1]
+            # Expecting format of "<xref uid="uid">text</xref>"
+            if front_tag in summary_part:
+                # Retrieve the index for starting position of xref tag
+                initial_index += summary_part.find(front_tag)
+
+                # Find the index of the end of xref tag, relative to the start of xref tag
+                end_tag_index = initial_index + parsed_text[initial_index:].find(end_tag) + end_len
+
+                # Retrieve the entire xref tag
+                original_type = parsed_text[initial_index:end_tag_index]
                 initial_index += len(original_type)
                 original_type = " ".join(filter(None, re.split(r'\n|  |\|\s|\t', original_type)))
-                safe_type = 'xref_' + original_type[6:-1]
+
+                # Extract text from "<xref uid="uid">text</xref>"
+                index = original_type.find(">")
+                safe_type = 'xref_' + original_type[index+1:index+(original_type[index:].find("<"))]
             else:
                 raise ValueError("Encountered unexpected type in Exception docstring.")
 
@@ -451,10 +496,20 @@ def _extract_docstring_info(summary_info, summary, name):
             cur_type = word
             if cur_type in [':type', ':param', ':raises', ':raises:']:
                 index += 1
-                arg_name = parsed_text[index][:-1]
-                # Initialize empty dictionary if it doesn't exist already
-                if arg_name not in summary_info[var_types[cur_type]] and ':raises' not in cur_type:
-                    summary_info[var_types[cur_type]][arg_name] = {}
+                # Exception that's not xref should be treated same as other names
+                if ':raises' not in cur_type or 'xref' not in parsed_text[index]:
+                    arg_name = parsed_text[index][:-1]
+                # xrefs are treated by taking its second half and combining the two
+                elif ':raises' in cur_type and 'xref' in parsed_text[index]:
+                    arg_name = f'{parsed_text[index]} {parsed_text[index+1][:-1]}'
+                    index += 1
+
+                try:
+                    # Initialize empty dictionary if it doesn't exist already
+                    if arg_name not in summary_info[var_types[cur_type]] and ':raises' not in cur_type:
+                        summary_info[var_types[cur_type]][arg_name] = {}
+                except KeyError:
+                    raise KeyError(f"Encountered wrong formatting, please check docstring for {name}")
 
             # Empty target string
             words = []
@@ -494,22 +549,55 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
     try:
         if _type in [METHOD, FUNCTION]:
             argspec = inspect.getfullargspec(obj) # noqa
+            type_map = {}
+            if argspec.annotations:
+                for annotation in argspec.annotations:
+                    if annotation == "return":
+                        continue
+                    # Extract names for simple types.
+                    try:
+                        type_map[annotation] = (argspec.annotations[annotation]).__name__
+                    # Try to extract names for more complicated types.
+                    except AttributeError:
+                        vartype = argspec.annotations[annotation]
+                        try:
+                            type_map[annotation] = str(vartype._name)
+                            if vartype.__args__:
+                                type_map[annotation] += str(vartype.__args__)[:-2] + ")"
+                        except AttributeError:
+                            print(f"Could not parse argument information for {annotation}.")
+                            continue
+
             for arg in argspec.args:
+                arg_map = {}
                 # Ignore adding in entry for "self"
                 if arg != 'cls':
-                    args.append({'id': arg})
+                    arg_map['id'] = arg
+                    if arg in type_map:
+                        arg_map['var_type'] = type_map[arg]
+                        args.append(arg_map)
             if argspec.varargs:
                 args.append({'id': argspec.varargs})
             if argspec.varkw:
                 args.append({'id': argspec.varkw})
+            # Try to add default values. Currently does not work if there is * argument present.
             if argspec.defaults:
-                for count, default in enumerate(argspec.defaults):
-                    cut_count = len(argspec.defaults)
-                    # Only add defaultValue when str(default) doesn't contain object address string(object at 0x)
-                    # inspect.getargspec method will return wrong defaults which contain object address for some default values, like sys.stdout
-                    # Match the defaults with the count
-                    if 'object at 0x' not in str(default):
-                        args[len(args) - cut_count + count]['defaultValue'] = str(default)
+                # Attempt to add default values to arguments.
+                try:
+                    for count, default in enumerate(argspec.defaults):
+                        # Find the first index which default arguments start at.
+                        # Every argument after this offset_count all have default values.
+                        offset_count = len(argspec.defaults)
+                        # Find the index of the current default value argument
+                        index = len(args) + count - offset_count
+
+                        # Only add defaultValue when str(default) doesn't contain object address string(object at 0x)
+                        # inspect.getargspec method will return wrong defaults which contain object address for some default values, like sys.stdout
+                        if 'object at 0x' not in str(default):
+                            args[index]['defaultValue'] = str(default)
+                # If we cannot find the argument, it is missing a type and was taken out intentionally.
+                except IndexError:
+                    pass
             try:
                 lines = inspect.getdoc(obj)
                 lines = lines.split("\n") if lines else []
@@ -547,11 +635,13 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
 
         # Get folder name from conf.py
         path = os.path.join(app.config.folder, path)
+        app.env.docfx_class_paths[cls] = path
 
         # append relative path defined in conf.py (in case of "binding python" project)
         try:
             source_prefix  # does source_prefix exist in the current namespace
             path = source_prefix + path
+            app.env.docfx_class_paths[cls] = path
         except NameError:
             pass
 
@@ -620,12 +710,26 @@ def _create_datam(app, cls, module, name, _type, obj, lines=None):
 
     if args:
         variables = summary_info['variables']
+        arg_id = []
         for arg in args:
+            arg_id.append(arg['id'])
+
             if arg['id'] in variables:
                 # Retrieve argument info from extracted map of variable info
                 arg_var = variables[arg['id']]
                 arg['var_type'] = arg_var.get('var_type') if arg_var.get('var_type') else ''
                 arg['description'] = arg_var.get('description') if arg_var.get('description') else ''
+
+        # Add any variables we might have missed from extraction.
+        for variable in variables:
+            if variable not in arg_id:
+                new_arg = {
+                    "id": variable,
+                    "var_type": variables[variable].get('var_type'),
+                    "description": variables[variable].get('description')
+                }
+                args.append(new_arg)
+
         datam['syntax']['parameters'] = args
 
     if sig:
@@ -845,20 +949,22 @@ def find_unique_name(package_name, entries):
         #   "google.cloud.spanner.v1.params_v1.types"
         #   "google.cloud.spanner.v1.instance_v1.types"
         # it will return "instace_v1" or "params_v1" and "types".
-        if name != "google" and name != "cloud" and entries[name] == 1:
+        # Also ensure that if name == package_name[-1], we only return one of
+        # the duplicate and not both.
+        if name != "google" and name != "cloud" and entries[name] == 1 and name != package_name[-1]:
             return [name, package_name[-1]]
 
-    # If there is no way to disambiguate, return the identifier name
+    # If there is no way to disambiguate or we found duplicates, return the identifier name.
     return [package_name[-1]]
 
 # Used to disambiguate names that have same entries.
 # Returns a dictionary of names that are disambiguated in the form of:
 # {uidname: disambiguated_name}
-def disambiguate_toc_name(toc_yaml):
+def disambiguate_toc_name(pkg_toc_yaml):
     name_entries = {}
     disambiguated_names = {}
 
-    for module in toc_yaml:
+    for module in pkg_toc_yaml:
         module_name = module['name']
         if module_name not in name_entries:
             name_entries[module_name] = {}
@@ -879,7 +985,7 @@ def disambiguate_toc_name(toc_yaml):
             # Update the dictionary of dismabiguated names
             disambiguated_names.update(disambiguate_toc_name(module['items']))
 
-    for module in toc_yaml:
+    for module in pkg_toc_yaml:
         module_name = module['name']
         # Check if there are multiple entires of module['name'], disambiguate if needed.
         if name_entries[module_name][module_name] > 1:
@@ -888,20 +994,111 @@ def disambiguate_toc_name(toc_yaml):
 
     return disambiguated_names
 
+
+# Combines pkg_toc_yaml entries with similar version headers.
+def group_by_package(pkg_toc_yaml):
+    new_pkg_toc_yaml = []
+    package_groups = {}
+    for module in pkg_toc_yaml:
+        package_group = find_package_group(module['uidname'])
+        if package_group not in package_groups:
+            package_name = pretty_package_name(package_group)
+            package_groups[package_group] = {
+                "name": package_name,
+                "uidname": package_group,
+                "items": []
+            }
+        package_groups[package_group]['items'].append(module)
+
+    for package_group in package_groups:
+        new_pkg_toc_yaml.append(package_groups[package_group])
+
+    return new_pkg_toc_yaml
+
+
+# Given the full uid, return the package group including its prefix.
+def find_package_group(uid):
+    return ".".join(uid.split(".")[:3])
+
+
+# Given the package group, make its name presentable.
+def pretty_package_name(package_group):
+    name = ""
+
+    # Retrieve only the package name
+    split_name = package_group.split(".")[-1]
+    # Capitalize the first letter of each package name part
+    capitalized_name = [part.capitalize() for part in split_name.split("_")]
+    return " ".join(capitalized_name)
+
+
+# For a given markdown file, extract its header line.
+def extract_header_from_markdown(mdfile_iterator):
+    for header_line in mdfile_iterator:
+        # Ignore licenses and other non-headers prior to the header.
+        if "#" in header_line:
+            break
+
+    if header_line.count("#") != 1:
+        raise ValueError(f"The first header of {mdfile_iterator.name} is not a h1 header: {header_line}")
+
+    # Extract the header name.
+    return header_line.strip("#").strip()
+
+
+# Given generated markdown files, incorporate them into the docfx_yaml output.
+# The markdown file metadata will be added to top level of the TOC.
+def find_markdown_pages(app, outdir):
+    # Use this to ignore markdown files that are unnecessary.
+    files_to_ignore = [
+        "index.md",     # merge index.md and README.md and index.yaml later.
+                        # See https://github.com/googleapis/sphinx-docfx-yaml/issues/105.
+
+        "reference.md", # Reference docs overlap with Overview. Will try and incorporate this in later.
+                        # See https://github.com/googleapis/sphinx-docfx-yaml/issues/106.
+
+        "readme.md",    # README does not seem to work in cloud site
+                        # See https://github.com/googleapis/sphinx-docfx-yaml/issues/107.
+
+        "upgrading.md", # Currently the formatting breaks, will need to come back to it.
+                        # See https://github.com/googleapis/sphinx-docfx-yaml/issues/108.
+    ]
+
+    markdown_dir = Path(app.builder.outdir).parent / "markdown"
+    if not markdown_dir.exists():
+        print("There's no markdown file to move.")
+        return
+
+    # For each file, if it is a markdown file move to the top level pages.
+    for mdfile in markdown_dir.iterdir():
+        if mdfile.is_file() and mdfile.name.lower() not in files_to_ignore:
+            shutil.copy(mdfile, f"{outdir}/{mdfile.name.lower()}")
+
+            # Extract the header name for TOC.
+            with open(mdfile) as mdfile_iterator:
+                name = extract_header_from_markdown(mdfile_iterator)
+
+            # Add the file to the TOC later.
+            app.env.markdown_pages.append({
+                'name': name,
+                'href': mdfile.name.lower(),
+            })
+
+
 def build_finished(app, exception):
     """
     Output YAML on the file system.
     """
 
     # Used to get rid of the uidname field for cleaner toc file.
-    def sanitize_uidname_field(toc_yaml):
-        for module in toc_yaml:
+    def sanitize_uidname_field(pkg_toc_yaml):
+        for module in pkg_toc_yaml:
             if 'items' in module:
                 sanitize_uidname_field(module['items'])
             module.pop('uidname')
 
-    def find_node_in_toc_tree(toc_yaml, to_add_node):
-        for module in toc_yaml:
+    def find_node_in_toc_tree(pkg_toc_yaml, to_add_node):
+        for module in pkg_toc_yaml:
             if module['uidname'] == to_add_node:
                 return module
 
@@ -916,6 +1113,7 @@ def build_finished(app, exception):
         if 'source' in obj and 'path' in obj['source'] and obj['source']['path']:
             if obj['source']['path'].endswith(INITPY):
                 obj['type'] = 'subPackage'
+                obj['summary'] = "API documentation for `{}` package.".format(obj['name'])
                 return
 
         for child_uid in obj['children']:
@@ -933,7 +1131,10 @@ def build_finished(app, exception):
     ))
     ensuredir(normalized_outdir)
 
-    toc_yaml = []
+    # Add markdown pages to the configured output directory.
+    find_markdown_pages(app, normalized_outdir)
+
+    pkg_toc_yaml = []
     # Used to record filenames dumped to avoid confliction
     # caused by Windows case insensitive file system
     file_name_set = set()
@@ -1071,54 +1272,56 @@ def build_finished(app, exception):
                         del(obj['source'])
 
             yaml_map[uid] = [yaml_data, references]
-           
+
             # Parse the name of the object.
             # Some types will need additional parsing to de-duplicate their names and contain
-            # a portion of their parent name for better disambiguation. This is done in 
+            # a portion of their parent name for better disambiguation. This is done in
             # disambiguate_toc_name
-            
-            node_name = obj.get('class').split(".")[-1] if obj.get('class') else obj['name']
+
+            node_name = uid.split(".")[-1]
 
             # Build nested TOC
             if uid.count('.') >= 1:
                 parent_level = '.'.join(uid.split('.')[:-1])
-                found_node = find_node_in_toc_tree(toc_yaml, parent_level)
+                found_node = find_node_in_toc_tree(pkg_toc_yaml, parent_level)
 
                 if found_node:
-                    found_node.pop('uid', 'No uid found')
                     found_node.setdefault(
-                      'items', 
+                      'items',
                       [{'name': 'Overview', 'uidname': parent_level, 'uid': parent_level}]
                     ).append({
                       'name': node_name,
-                      'uidname': uid, 
+                      'uidname': uid,
                       'uid': uid
                     })
                 else:
-                    toc_yaml.append({
-                      'name': node_name, 
-                      'uidname': uid, 
+                    pkg_toc_yaml.append({
+                      'name': node_name,
+                      'uidname': uid,
                       'uid': uid
                     })
 
             else:
-                toc_yaml.append({
-                  'name': node_name, 
-                  'uidname': uid, 
+                pkg_toc_yaml.append({
+                  'name': node_name,
+                  'uidname': uid,
                   'uid': uid
                 })
 
-    if len(toc_yaml) == 0:
+    # Exit if there are no generated YAML pages or Markdown pages.
+    if len(pkg_toc_yaml) == 0 and len(app.env.markdown_pages) == 0:
         raise RuntimeError("No documentation for this module.")
 
     # Perform additional disambiguation of the name
-    disambiguated_names = disambiguate_toc_name(toc_yaml)
+    disambiguated_names = disambiguate_toc_name(pkg_toc_yaml)
+
+    pkg_toc_yaml = group_by_package(pkg_toc_yaml)
 
     # Keeping uidname field carrys over onto the toc.yaml files, we need to
     # be keep using them but don't need them in the actual file
-    toc_yaml_with_uid = copy.deepcopy(toc_yaml)
+    pkg_toc_yaml_with_uid = copy.deepcopy(pkg_toc_yaml)
 
-    sanitize_uidname_field(toc_yaml)
+    sanitize_uidname_field(pkg_toc_yaml)
 
     toc_file = os.path.join(normalized_outdir, 'toc.yml')
     with open(toc_file, 'w') as writable:
@@ -1126,43 +1329,12 @@ def build_finished(app, exception):
             dump(
                 [{
                     'name': app.config.project,
-                    'items': [{'name': 'Overview', 'uid': 'project-' + app.config.project}] + toc_yaml
+                    'items': [{'name': 'Overview', 'uid': 'project-' + app.config.project}] + app.env.markdown_pages + pkg_toc_yaml
                 }],
                 default_flow_style=False,
             )
         )
 
-    index_file = os.path.join(normalized_outdir, 'index.yml')
-    index_children = []
-    index_references = []
-    for item in toc_yaml_with_uid:
-        index_children.append(item.get('uidname', ''))
-        index_references.append({
-            'uid': item.get('uidname', ''),
-            'name': item.get('name', ''),
-            'fullname': item.get('uidname', ''),
-            'isExternal': False
-        })
-    with open(index_file, 'w') as index_file_obj:
-        index_file_obj.write('### YamlMime:UniversalReference\n')
-        dump(
-            {
-                'items': [{
-                    'uid': 'project-' + app.config.project,
-                    'name': app.config.project,
-                    'fullName': app.config.project,
-                    'langs': ['python'],
-                    'type': 'package',
-                    'kind': 'distribution',
-                    'summary': 'Reference API documentation for `{}`.'.format(app.config.project),
-                    'children': index_children
-                }],
-                'references': index_references
-            },
-            index_file_obj,
-            default_flow_style=False
-        )
-
     # Output files
     for uid, data in iter(yaml_map.items()):
 
@@ -1174,7 +1346,7 @@ def build_finished(app, exception):
                     obj['name'] = disambiguated_names[obj_full_name]
                     if obj['type'] == 'subPackage':
                         obj['summary'] = "API documentation for `{}` package.".format(obj['name'])
-      
+
         # data is formatted as [yaml_data, references]
         yaml_data, references = data
 
@@ -1205,6 +1377,38 @@ def build_finished(app, exception):
 
         file_name_set.add(filename)
 
+    index_file = os.path.join(normalized_outdir, 'index.yml')
+    index_children = []
+    index_references = []
+    for package in pkg_toc_yaml_with_uid:
+        for item in package.get("items"):
+            index_children.append(item.get('uidname', ''))
+            index_references.append({
+                'uid': item.get('uidname', ''),
+                'name': item.get('name', ''),
+                'fullname': item.get('uidname', ''),
+                'isExternal': False
+            })
+    with open(index_file, 'w') as index_file_obj:
+        index_file_obj.write('### YamlMime:UniversalReference\n')
+        dump(
+            {
+                'items': [{
+                    'uid': 'project-' + app.config.project,
+                    'name': app.config.project,
+                    'fullName': app.config.project,
+                    'langs': ['python'],
+                    'type': 'package',
+                    'kind': 'distribution',
+                    'summary': 'Reference API documentation for `{}`.'.format(app.config.project),
+                    'children': index_children
+                }],
+                'references': index_references
+            },
+            index_file_obj,
+            default_flow_style=False
+        )
+
     '''
     # TODO: handle xref for other products.
     xref_file = os.path.join(normalized_outdir, 'xrefs.yml')
